#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Chris Lumens <clumens@redhat.com>
#
import atexit
import glob
import imputil
import logging
import os
from random import randint
import shutil
import signal
import sys

def cleanup(pid):
    # Kill the results server so we don't have it still running after we're
    # gone, blocking up the socket.
    os.kill(pid, signal.SIGTERM)

def resultsServer(resultsdir, serverToken):
    from SimpleXMLRPCServer import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
    from xmlrpclib import Fault

    def _ensureDestDir(resultsdir, testresultsdir):
        destDir = os.path.normpath("%s/%s" % (resultsdir, testresultsdir))
        if not destDir.startswith(resultsdir):
            return None

        if not os.path.exists(destDir):
            os.makedirs(destDir)

        return destDir

    # This method provides a way for the guest to indicate what the result
    # of the test case was - something like "PASSED" or "FAILED".
    def _writeStatus(token, testresultsdir, status):
        if token != serverToken:
            return Fault("Authentication error", "Invalid token provided")

        destDir = _ensureDestDir(resultsdir, testresultsdir)
        if not destDir:
            return Fault("Could not store file", "Bad destination directory")

        # Just touch an empty file.
        open("%s/%s" % (destDir, status), "w").close()

        # Also log the result
        logging.info("Test %s: %s" % (testresultsdir, status))
        return ""

    # This method provides a way for the guest to store one file on the host.
    def upload(token, testresultsdir, filename, contents):
        if token != serverToken:
            return Fault("Authentication error", "Invalid token provided")

        # If the destination doesn't already exist (perhaps because this is
        # the first file coming in for a given test), create it.  Need to be
        # careful here.
        if filename.find("/") != -1:
            return Fault("Could not store file", "Bad filename given: %s" % filename)

        destDir = _ensureDestDir(resultsdir, testresultsdir)
        if not destDir:
            return Fault("Could not store file", "Bad destination directory")

        with open("%s/%s" % (destDir, filename), "wb") as fo:
            fo.write(contents.data)

        return ""

    server = SimpleXMLRPCServer(("localhost", 4747), SimpleXMLRPCRequestHandler, False)

    server.register_introspection_functions()
    server.register_function(lambda token, testresultsdir: _writeStatus(token, testresultsdir, "FAILED"),
                             "failed")
    server.register_function(lambda token, testresultsdir: _writeStatus(token, testresultsdir, "PASSED"),
                             "passed")
    server.register_function(upload)

    server.serve_forever()

class StorageTestFramework(object):
    @property
    def livecd(self):
        if self._livecd:
            return self._livecd

        filelist = glob.glob(self.livecddir + "/anaconda-storage.iso")
        if not filelist:
            return None
        else:
            return filelist[0]

    @property
    def suites(self):
        # Return a list of all the *Suite classes in suites/*.py.
        if self._suites:
            return self._suites

        tstList = set()

        # Make sure self.suitedir is in the system path so imputil works.
        if not self.suitedir in sys.path:
            sys.path.insert(0, self.suitedir)
            sys.path.insert(0, os.path.dirname(self.suitedir))

        # Get a list of all *.py files in self.suitedir
        moduleList = []
        lst = map(lambda x: os.path.splitext(os.path.basename(x))[0],
                  glob.glob(self.suitedir + "/*.py"))

        # Inspect each .py file found.
        for module in lst:
            if module in ["__init__"]:
                logging.debug("Skipping suite '%s'" % module)
                continue

            # Attempt to load the found module.
            try:
                found = imputil.imp.find_module(module)
                loaded = imputil.imp.load_module(module, found[0], found[1], found[2])
            except ImportError, e:
                logging.debug("ImportError: Failed to import suite '%s'\n%s" % (module, e))
                continue

            # Find class names that match the supplied pattern.
            for obj in loaded.__dict__.keys():
                if obj.endswith("TestSuite"):
                    tstList.add(loaded.__dict__[obj])

        self._suites = list(tstList)
        return self._suites

    @property
    def token(self):
        # Create a simple authentication token for the results server.  This will
        # get passed to each guest via substitution into driver.py, and the guest
        # must send the token back when asking to store a result.  This doesn't
        # prevent everything bad from happening, but should make it significantly
        # harder for random outsiders to log results.
        if self._token:
            return self._token

        # Taken from mimimon:
        # http://git.fedorahosted.org/git/?p=autoqa.git;a=blob;f=tests/rats_install/minimon.py
        token_char = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
        length = 10

        self._token = "".join([token_char[randint(0,63)] for i in range(0,length)])
        return self._token

    def _createLiveCD(self):
        live_ksfile = os.path.join(self.resultsdir, "ks.cfg")
        # If a kickstart was provided, use it
        if self.kickstart and os.path.isfile(self.kickstart):
            ks_template = self.kickstart
        # Otherise no kickstart file was provided, build one using the template
        else:
            ks_template = self.basedir + "/storage-test.ks"

        if not os.path.isfile(ks_template):
            raise Exception("Unable to file kickstart file '%s'" % ks_template)

        # Create a new kickstart using the provided template
        dst = open(live_ksfile, "a")
        for url in self.repos:
            dst.write("repo --name=repo-%s --baseurl=\"%s\"\n" % (self.repos.index(url), url))

        # Reuse existing template
        src = open(ks_template, 'r')
        data = src.read()
        while data:
            dst.write(data)
            data = src.read()
        src.close()
        dst.close()

        logging.info("----- Creating Live CD image -----")

        # Create the live image
        # FIXME - add --quiet option when patch is accepted upstream (see
        # http://lists.fedoraproject.org/pipermail/livecd/2010-November/006518.html)
        cmd = "livecd-creator -c %s -f anaconda-storage --cache=$HOME/livecd-cache " \
              "--debug --logfile=%s/livecd-creator.log" % (live_ksfile, self.resultsdir)
        rc = os.system(cmd)
        if rc:
            return rc # failure

        if self.livecddir != os.getcwd():
            shutil.move("anaconda-storage.iso", self.livecddir)

        return 0 # pass

    def run(self, args=None):
        # Display test suite configuration
        logging.info("----- Suite Configuration -----")
        for k in ["basedir", "tmpdir", "diskimagesdir", "kickstart", \
                  "livecddir", "repos", "resultsdir", "suitedir"]:
            logging.info("%s: %s" % (k, getattr(self, k, "")))

        # Run all the test suites.  We don't need to worry about waiting for the
        # results server to start up since we're going to be doing long things like
        # booting VMs here.
        suites = self.suites

        if len(suites) <= 0:
            logging.critical("No testsuites found in '%s'" % self.suitedir)
            sys.exit(1)

        if not self.livecd and self._createLiveCD():
            logging.critical("Failed to make the live CD.  Exiting")
            sys.exit(1)

        for suite in suites:
            # If we were given test suites to run on the command line, only
            # run those.
            if args and suite.name not in args:
                logging.info("Skipping '%s' test" % suite.name)
                continue

            oldwd = os.getcwd()
            os.chdir(self.basedir)

            obj = suite(self.basedir, self.diskimagesdir)
            if not obj.runnable:
                continue

            logging.info("===== Running Test Suite %s =====" % obj.name)
            logging.info("----- Creating Disk Images -----")
            obj.setup()
            obj.run(self.token, self.livecd, self.resultsdir)
            obj.cleanup()

            os.chdir(oldwd)

        # Now that we're all done, do a final check of what passed and failed
        # and print a report.
        self.showResults()

    def startReportServer(self):
        # Fork off the XML-RPC server to accept results from the test cases.
        self._server_pid = os.fork()
        if self._server_pid:
            atexit.register(cleanup, self._server_pid)
        else:
            resultsServer(self.resultsdir, self.token)

    def stopReportServer(self):
        cleanup(self._server_pid)

    def setupDirs(self):
        # Make sure there's no old disk images laying around.
        if os.path.exists(self.diskimagesdir):
            shutil.rmtree(self.diskimagesdir)

        os.makedirs(self.diskimagesdir)

        # Would be nice to leave the results directory around for later use, though.
        if not os.path.exists(self.resultsdir):
            os.makedirs(self.resultsdir)

    def showResults(self):
        passedTests = 0
        failedTests = 0
        passedSuites = 0
        failedSuites = 0

        logging.info("===== Storage Test Results =====")

        for suite in glob.glob(self.resultsdir + "/*"):
            if not os.path.isdir(suite):
               continue

            thisSuitePassed = True
            name = os.path.basename(suite)

            for test in glob.glob(suite + "/*"):
                if os.path.exists(test + "/PASSED"):
                    passedTests += 1
                else:
                    thisSuitePassed = False
                    failedTests += 1

            if thisSuitePassed:
                logging.info("Suite %s:  PASSED" % name)
                passedSuites += 1
            else:
                logging.info("Suite %s:  FAILED" % name)
                failedSuites += 1

        logging.info("Tests Passed:  %s" % passedTests)
        logging.info("Tests Failed:  %s" % failedTests)
        logging.info("Suites Passed: %s" % passedSuites)
        logging.info("Suites Failed: %s" % failedSuites)

        # And finally, touch a file giving the overall status of the entire
        # storage test suite.  This way, any wrapped (like autoqa) only has
        # to check the existance of one file instead of doing any analysis.
        if failedSuites == 0:
            open("%s/PASSED" % self.resultsdir, "w").close()
        else:
            open("%s/FAILED" % self.resultsdir, "w").close()

    def __init__(self, *args, **kwargs):
        self.basedir = kwargs.get("basedir", os.getcwd())
        self.tmpdir = kwargs.get("tmpdir", "/tmp")

        self.diskimagesdir = kwargs.get("diskimagesdir", self.tmpdir + "/images")
        self.kickstart = kwargs.get("kickstart", "")
        self.livecddir = kwargs.get("livecddir", self.basedir)
        self.repos = kwargs.get("repos", [])
        self.resultsdir = kwargs.get("resultsdir", self.tmpdir + "/results")
        self.suitedir = kwargs.get("suitedir", self.basedir + "/suites")

        self._livecd = None
        self._suites = []
        self._token = None
