#!/usr/bin/python
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Chris Lumens <clumens@redhat.com>

import sys

from cases import *

class LVMonRAID0Runner(BaseRunner):
    def writeKickstart(self):
        fo = open("/tmp/kickstart.ks", "w")
        fo.write("clearpart --all\n")
        fo.write("zerombr\n")
        fo.write("part /boot --fstype=ext4 --size=500\n")
        fo.write("part swap --fstype=swap --recommended\n")
        fo.write("part raid.01 --size=2048\n")
        fo.write("part raid.02 --size=2048\n")
        fo.write("part pv.02 --size=2048\n")
        fo.write("raid pv.01 --device=0 --fstype=ext4 --level=RAID0 raid.01 raid.02\n")
        fo.write("volgroup LVMONRAID --pesize=32768 pv.01 pv.02\n")
        fo.write("logvol / --fstype=ext4 --grow --size=2048 --name=root --vgname=LVMONRAID\n")
        fo.close()

class LVMonRAID0TestCase(BaseTestCase):
    """Verify we have the following:

       - 500 MB /boot partition
       - swap partition with size inside our recommendations
       - two equally sized partitions with a RAID format
       - a third partition with a PV format
       - a VG on md0
       - / on LVM
    """
    def test_boot(self):
        self.default_test_boot()

    def test_lvm(self):
        # Check that we created a single raid device and gave it a PV format.
        raid = self.single(self.getDevs(self.state, type="mdarray"))
        self.failIf(not raid,
                    "Partitioning did not create md0.")
        self.assertEqual(raid["format"]["type"], "lvmpv",
                         "A physical volume was not created on md0.")

        # Check that two PVs were created.  One's the RAID device from above,
        # leaving just the extra partition.
        pvs = self.getDevs(self.state, format="lvmpv")
        self.assert_(len(pvs) == 2,
                     "Partitioning did not create exactly two PVs.")

        # Check that a single VG was created and that it is composed of two
        # devices.
        vg = self.single(self.getDevs(self.state, type="lvmvg"))
        self.failIf(not vg,
                    "Partitioning did not create a volume group.")
        self.assert_(len(vg["parents"]) == 2,
                     "Volume group is not composed of exactly two devices.")

    def test_raid(self):
        # Check that we created two equal sized partitions of type mdmember.
        members = self.getDevs(self.state, format="mdmember")
        self.assert_(len(members) == 2,
                     "Partitioning did not create exactly two RAID members.")
        self.assert_(members[0]["length"] == members[1]["length"],
                     "Partitioning did not create equal sized RAID members.")

        # Check that we created a RAID0 device.
        raid = self.single(self.getDevs(self.state, type="mdarray"))
        self.assert_(raid["level"] == 0,
                     "Device is not RAID0.")
        self.assert_(raid["name"] == "md0",
                     "Partitioning did not create md0.")

    def test_root(self):
        rootDev = self.single(self.getDevs(self.state, mountpoint="/"))
        self.failIf(not rootDev,
                    "Partitioning did not create /.")

        # Check that / was created on top of an LV.
        self.assert_(rootDev["format"]["type"] == "ext4",
                     "Type of / is not ext4.")
        self.assert_(rootDev["type"] == "lvmlv",
                     "/ was not created on a logical volume.")

    def test_swap(self):
        from pyanaconda.iutil import swapSuggestion

        swapDev = self.single(self.getDevs(self.state, format="swap"))
        self.failIf(not swapDev,
                    "autopart did not create swap.")

        # Swap size can vary wildly, so just check that swap exists and is inside
        # the range anaconda suggests.
        (minSwap, maxSwap) = swapSuggestion()
        self.assert_(minSwap <= swapDev["size"] <= maxSwap,
                     "swap is not within suggested range")

        # Check the fs type and underlying device of swap.
        self.assert_(swapDev["format"]["type"] == "swap",
                     "Type of swap is not swap.")
        self.assertTrue(self.onPartition(swapDev),
                        "swap was not created on a partition.")
