#!/usr/bin/python
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Chris Lumens <clumens@redhat.com>
"""
Base classes for creating test cases.  A storage test case consists of two
classes, all of which must be provided no matter how simple the test case:

    (1) A subclass of BaseRunner, which creates a kickstart file and runs it
        through anaconda's storage code.

    (2) A subclass of BaseTestCase, which validates that the final partitioning
        state is as expected.

A test case itself is a single file with these classes in it.  The test case
should have some short but useful name.  Then, this name should be used as the
file name as well as the initial part of all three of the above classes.  For
instance, if you are writing a test case for making sure that /usr on RAID5
works, you should use a name like "UsrOnRAID5.".  Then, your file would be
UsrOnRAID5.py and you would create classes named UsrOnRAID5Creator,
UsrOnRAID5Runner, and UsrOnRAID5TestCase.

While picky, these names are very important.  The test case framework searches
for objects with these naming conventions.  If not found, your test case will
not be executed.

Since the parts of a test case are just objects, it's easy to construct a test
case built on top of another one via inheritance.  In this case, almost all the
work will be in the TestCase class.  All tests all the way up the inheritance
chain will be run, not just those in the new class.
"""

from ConfigParser import ConfigParser
import glob
from optparse import OptionParser
import os
import parted
import shelve
import shutil
import subprocess
import sys
import tempfile
import unittest
import xmlrpclib

def report(server, token, reportdir, wasSuccessful):
    # Send a file to the results server.
    def _sendFile(server, token, reportdir, filename):
        if not os.path.exists(filename):
            return

        with open(filename) as fo:
            contents = "".join(fo.readlines())
            server.upload(token, reportdir, os.path.basename(filename),
                          xmlrpclib.Binary(contents))

    # Send a quick and simple pass vs. fail message so the server
    # doesn't have to analyze the log files to determine whether the test
    # passed.
    if wasSuccessful:
        server.passed(token, reportdir)
    else:
        server.failed(token, reportdir)

    # And then upload a bunch of files so we have more information in
    # the event of a failure.
    _sendFile(server, token, reportdir, "/tmp/anaconda.log")
    _sendFile(server, token, reportdir, "/tmp/program.log")
    _sendFile(server, token, reportdir, "/tmp/storage.log")
    _sendFile(server, token, reportdir, "/tmp/storage.state")
    _sendFile(server, token, reportdir, "/tmp/syslog")
    _sendFile(server, token, reportdir, "/var/log/boot.log")

    for f in glob.glob("/tmp/anaconda-tb-*"):
        _sendFile(server, token, reportdir, f)

class Creator:
    """A BaseCreator instance is one of three parts of a test case.  It runs
       on the host and sets up a VM to run the rest of the test case in.  It
       creates disk images, initializes them if necessary, and performs
       cleanup actions after the test is complete.

       All test cases must provide a setup method and a run method.  They
       may optionally provide an __init__ and cleanup if doing something
       unusual.
    """
    def __init__(self, name, imagesdir, drives, *args, **kwargs):
        """Create a new BaseCreator instance.  As part of this, any temporary
           directories needed to store disk images will also be created.  Test
           cases should not normally need to supply this method.

           Required arguments:

           name -- The name of this test case.
           imagesdir -- The directory to use when creating disk images.
           drives -- A list of paths to disk images that have previously been
                     created with qemu-img (or some other tool).

           Arguments that may be passed via kwargs:

           reqMemory -- The amount of memory the VM to run this test case
                        needs in MB, or 512 by default.
           runnerClassAttrs -- A dict of attributes and values that will
                               be added to the RunnerClass that will be
                               executed by this test suite.
           testClassAttrs -- The same as runnerClassAttrs, but for TestClass.
                             Most of the time, these two dicts will probably
                             need to be identical.
        """
        self.drives = drives
        self.imagesdir = imagesdir
        self.name = name

        self.reqMemory = kwargs.get("reqMemory", 512)
        self.runnerClassAttrs = kwargs.get("runnerClassAttrs", {})
        self.testClassAttrs = kwargs.get("testClassAttrs", {})

        self._proc = None

    def create(self):
        # Test cases can inherit from each other.  Being clever and only putting
        # into the ISO image what's required for the test case would mean figuring
        # out module dependencies here.  Instead we'll just copy everything in and
        # use a parameter to tell which test should be run.
        os.system("mkisofs -quiet -v -o %s/%s.iso -R -V 'ANACTEST' -m '*.iso' \
                   -m '*.ks' -m runtest.py -m 'docs/*' \
                   . 2>&1" %
                  (self.imagesdir, self.name))

    def die(self):
        """If a qemu process is currently running, kill it immediately."""
        if self._proc:
            self._proc.kill()
            self._proc = None

    def run(self, livecd):
        """Execute this test case.  This is done by running qemu-kvm against
           whatever disk images the setup() method put into self.drives.  This
           method should not be overridden by test cases but if you must, at
           least call BaseCreator.run after doing whatever other setup is
           required.  This saves having to duplicate the qemu argument list.
        """
        args = ["/usr/bin/qemu-kvm",
                "-m", str(self.reqMemory),
                "-boot", "d",
                "-drive", "file=%s,media=cdrom" % livecd,
                "-drive", "file=%s/%s.iso,media=cdrom" % (self.imagesdir, self.name)]

        for drive in self.drives:
            args += ["-drive", "file=%s,media=disk" % drive]

        # FIXME - this assumes that DISPLAY=:9 is available
        if not os.environ.has_key("DISPLAY"):
            args += ["-vnc", ":9"]

        # Save a reference to the running qemu process so we can later kill
        # it if necessary.  For now, the only reason we'd want to kill it is
        # an expired timer.
        self._proc = subprocess.Popen(args)
        self._proc.wait()
        self._proc = None

class BaseRunner:
    """ A BaseRunner instance is one of three parts of a test case.  It runs
        in the VM and runs a kickstart file through anaconda's storage code,
        reports results, and handles exceptions within anaconda to make sure
        they are also reported.

        All test cases must provide an instance of this class, but for the
        most part test cases will never need to do anything other than provide
        a writeKickstart method.  On rare occassion (such as running one test
        case as the input to a second), it may also be helpful to override
        the runAnaconda method.  However, those must still call
        BaseRunner.runAnaconda.  They may call it multiple times, though.
    """
    def __init__(self, driver):
        """Create a new BaseRunner instance.

           Class attributes:

        """
        self.driver = driver

    def runAnaconda(self):
        """Given a pre-existing /tmp/kickstart.ks file, run it through
           anaconda's storage code making sure to handle any exceptions
           that turn up.
        """
        import pyanaconda
        from pyanaconda import kickstart
        from pyanaconda.storage import storageInitialize, storageComplete
        from pyanaconda.storage.partitioning import doAutoPartition
        from pyanaconda.packages import turnOnFilesystems

        from intf import TestInstallInterface
        from exception import initExceptionHandling

        self.ana = pyanaconda.Anaconda()
        self.ana.displayMode = "c"
        self.ana.intf = TestInstallInterface()

        # Make sure we get /tmp/anaconda-tb-* on error - this is as early as this can
        # be set up.
        reportCB = lambda: report(xmlrpclib.ServerProxy("http://%s" % self.driver.server),
                                  self.driver.token,
                                  self.driver.reportdir,
                                  False)
        self.ana.mehConfig = initExceptionHandling(self.ana.intf, self.ana,
                                                   reportCB)

        # Process the kickstart file.
        self.writeKickstart()
        kickstart.preScriptPass(self.ana, "/tmp/kickstart.ks")
        self.ana.ksdata = kickstart.parseKickstart(self.ana, "/tmp/kickstart.ks")

        # Set up default partitioning, etc.
        self.ana.instClass.configure(self.ana)

        # Now apply the kickstart file and partitioning settings.
        storageInitialize(self.ana)
        self.ana.ksdata.execute()
        doAutoPartition(self.ana)
        storageComplete(self.ana)
        turnOnFilesystems(self.ana)
        self.ana.storage.write("/")

    def report(self, testClass):
        """Report test case results to the server running on the host.  This
           method is run after runAnaconda has been executed.  It also does
           some simple cleanup tasks.  There should be little reason for
           subclasses to override this method.
        """
        # Load and run the tests for this case.
        tests = unittest.TestLoader().loadTestsFromTestCase(testClass)
        results = unittest.TextTestRunner().run(tests)

        s = xmlrpclib.ServerProxy("http://%s" % self.driver.server)
        report(s,
               self.driver.token,
               self.driver.reportdir,
               results.wasSuccessful())

        # Clean up for the next test case.
        self.ana.storage.umountFilesystems()

    def setup(self):
        """This method performs any disk configuration required of the test
           case that must be run inside the VM.  This is needed because disk
           images are constructed wth qemu-img as QCOW2 images, which are
           initially very small and only grow as needed when running inside a
           VM.  Therefore, they cannot be used with parted or other tools.
           This method runs before writeKickstart or runAnaconda.
        """
        return

    def writeKickstart(self):
        """At its heart, a test case is really just running the partitioning
           section of a kickstart file through anaconda's storage code and
           validating that we got what we asked for.  This method must write
           out the kickstart snippet to /tmp/kickstart.ks so the rest of
           the framework may act.
        """
        if self.__class__ is BaseRunner:
            raise NotImplementedError, "writeKickstart() not implemented for BaseRunner"

class BaseTestCase(unittest.TestCase):
    """A BaseTestCase instance is one of three parts of a test case.  It runs
       on the VM and validates that anaconda created the partitioning as
       expected.  This is the most significant chunk of test case writing and
       will take the most time.

       This class is a wrapper around Python's unittest framework and provides
       all the test case functions necessary.  In addition to all the assert*
       and fail* methods, this class also provides a variety of helper library
       functions to make querying for information easier.

       Test validators must implement their individual tests as a set of
       methods following the test_* naming scheme, as python will look for
       these names.  Each individual test method should test one very small
       piece of the final partitioning state and call the appropriate
       fail*/pass*/assert* methods.

       In addition, BaseTestCase provides a set of default_test_* methods
       that may be useful.  If you want to make use of any of them, create
       a test_* method that calls default_test_*.
    """

    def setUp(self):
        """This method sets up self.state, through which all test methods
           query the final partitioning state.  Test cases should not need
           to override this method.  It is automatically called by Python.
        """
        self.shelf = shelve.open("/tmp/storage.state")

        # First, find the key for the final state of the disks.  If there is no
        # final key, installation aborted before it could be written out.
        self.state = None
        for k in self.shelf.keys():
            if k.endswith("final"):
                self.state = self.shelf[k]
                break

        if not self.state:
            raise RuntimeError, "No final storage state found"

    def tearDown(self):
        """This method cleans up after a test method has been run.  Test cases
           should not need to override this method.  It is automatically
           called by Python.
        """
        self.shelf.close()

    ###
    ### Library functions to help writing test cases easier.
    ###

    def disks(self, lst):
        """Return a list of all the disks in the storage shelf."""
        return self.getDevs(lst, type="disk")

    # Return all devices meeting a given criteria.
    def getDevs(self, lst, format=None, mountpoint=None, type=None):
        """Return all devices meeting a given criteria.  Only one criterium
           will be used per call.  If more than one is provided, the first
           in the method's argument list will take precedence.
        """
        if format:
            p = lambda d: d["format"]["type"] == format
        elif mountpoint:
            p = lambda d: d["format"].get("mountpoint", None) == mountpoint
        elif type:
            p = lambda d: d["type"] == type
        else:
            return []

        return filter(p, lst)

    # Verify that a list is a singleton, returning the element.
    def single(self, lst):
        """Verify that a given list is a singleton.  If so, return the one
           element.  If not, return None.  This is most useful when paired
           with getDevs.  For instance, if you query for a device with the
           mountpoint of /, you expect to only get a single disk.  However,
           getDevs returns a list.  This function wraps that test so the only
           possible return values are the element or None.
        """
        if len(lst) == 1:
            return lst[0]
        else:
            return None

    def onPartition(self, dev):
        """Does a given device exist on a physical partition?"""
        return dev["type"] in [parted.PARTITION_NORMAL, parted.PARTITION_LOGICAL]


    ###
    ### DEFAULT TESTS
    ### These should be called from within test_ functions in your *TestCase
    ### class.  If they were named as such here, they would always get called
    ### and that's not wanted.  Not every test case will set up the same
    ### partitioning.
    ###

    # Boot is almost always on a partition, so this is safe as a default test.
    def default_test_boot(self, fstype="ext4"):
        bootDev = self.single(self.getDevs(self.state, mountpoint="/boot"))
        self.failIf(not bootDev,
                    "autopart did not create /boot.")

        self.assertEqual(bootDev["format"]["type"], fstype,
                         "Type of /boot is not %s." % fstype)
        self.assertEqual(bootDev["size"], 500,
                         "Size of /boot does not match expected.")
        self.assertTrue(self.onPartition(bootDev),
                        "/boot was not created on a partition.")
