#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Josef Skladanka <jskladan@redhat.com>

import os
import shutil
import re
import autoqa.koji_utils
import autoqa.util as util
from autotest_lib.client.bin import utils
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher

class initscripts(AutoQATest):
    version = 1 # increment if setup() changes


    @ExceptionCatcher()
    def setup(self, **kwargs):
        utils.system('yum -y install beakerlib')


    @ExceptionCatcher()
    def initialize(self, config, nvr, **kwargs):
        super(self.__class__, self).initialize(config)

        self.detail.id = nvr
        self.koji = autoqa.koji_utils.SimpleKojiClientSession()
        self.koji.check_connection()


    def run_beakerlib_test(self, name, path_to_runtest):
        """
        there are lines like
          @echo "Requires:        openssh" >> $(METADATA)
        in the Makefile, specifying which packages are needed to run the respective test.
        this code extracts package names from these lines and installs the packages

        We also want to skip the "Requires" line with the package name, since it was installed
         from koji. Makefile has "RunFor" line, which specifies this package name.

        Returns: (result, outputs)
        """
        #install required packages
        makefile = os.path.join(path_to_runtest, "Makefile")
        if (os.path.isfile(makefile)):
            #grab the Requires lines
            req_lines = filter(lambda x: x.find("Requires:") > -1, open(makefile, 'r').readlines())
            requires = map(lambda x: re.findall(".*Requires:\s*([^\"']+).*", x)[0], req_lines)

            #grab the RunFor lines
            run_for_lines = filter(lambda x: x.find("RunFor:") > -1, open(makefile, 'r').readlines())
            run_for = map(lambda x: re.findall(".*RunFor:\s*([^\"']+).*", x)[0], run_for_lines)

            #filter out RunFor from Requires, as RunFor was installed from koji
            requires = filter(lambda x: x not in run_for, requires)
            if len(requires) > 0:
                utils.system("yum -y install %s" % " ".join(requires))

        #run the test
        cmd = os.path.join(path_to_runtest, "runtest.sh")
        cmd_result = utils.run(cmd, ignore_status=True,
                     stdout_tee=utils.TEE_TO_LOGS, stderr_tee=utils.TEE_TO_LOGS)
        result = 'PASSED' if cmd_result.exit_status == 0 else 'FAILED'
        outputs = cmd_result.stdout

        #store the log journal created by beakerlib
        """
        filters lines containing information about journal files (logs from the test script)
        :: [15:22:06] ::  JOURNAL XML: /tmp/beakerlib-CoEg7JM/journal.xml
        :: [15:22:06] ::  JOURNAL TXT: /tmp/beakerlib-CoEg7JM/journal.txt
        extracts paths to these files, ands stores the files in self.resultsdir
        """
        journal_lines = filter(lambda x: x.find("JOURNAL") > -1, outputs.splitlines())
        journal_files = map(lambda x: x.split(':')[-1].strip(), journal_lines)
        for fname in journal_files:
            shutil.copy(fname, os.path.join(self.resultsdir, "journal_%s.%s" % (name, fname.split('.')[-1])))

        return (result, outputs)


    @ExceptionCatcher()
    def run_once(self, kojitag, nvr, name, **kwargs):
        super(self.__class__, self).run_once()

        self.results_count = {'PASSED': 0, 'FAILED': 0} # count how many of which results do we have
        self.tests_passed = []
        self.tests_failed = []

        # add header
        self.log('%s\n%s\n%s' % ('='*40, nvr, '='*40))

        # find all tests for package $name
        testdir = os.path.join(self.bindir, "./tests/%s" % name)
        assert os.path.isdir(testdir), ('No initscript checker found for '
                                        'package %s' % name)
        tests = []
        """
        walks through ./tests/$NAME and all it's subdirs. Every time it finds
        file runtest.sh, it stores a tuple(test_name, path) to tests[],
        where test_name is 'openssh' for ./tests/openssh, openssh_subdir for ./tests/openssh/subdir, etc.
        """
        for (path, dirs, files) in os.walk(testdir):
            if "runtest.sh" in files:
                test_name = path.replace(os.path.join(self.bindir,
                            "./tests/"), "").replace(os.path.sep, "_")
                tests.append((test_name, path))

        # download packages from koji
        rpms = self.koji.get_nvr_rpms(nvr, rpm_dir=self.tmpdir,
            arches=[util.get_basearch(), 'noarch'])

        # and install them
        cmd = "yum -y --nogpgcheck localinstall %s" % " ".join(rpms)
        utils.system(cmd)

        #sort tests alphabetically by test_name
        tests.sort(key = lambda x: x[0])

        # run tests
        for test in tests:
            self.log("RUNNING: %s\n%s" % (test[0], '='*20))

            (result, outputs) = self.run_beakerlib_test(test[0], test[1])
            self.results_count[result] += 1
            self.log(outputs, printout=False)

            if result == 'PASSED':
                self.tests_passed.append(test[0])
            else:
                self.tests_failed.append(test[0])
            self.detail.update_result(result)

        summary_detail = {}
        summary_detail['PASSED'] = self.tests_passed
        summary_detail['FAILED'] = self.tests_failed
        self.detail.log.summary_detail = summary_detail

        # submit results, including email opt-in
        self.post_results(optin={'pkg': name, 'kojitag': kojitag})
