#!/usr/bin/python

# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Vojtech Aschenbrenner <vaschenb@redhat.com>

import operator
import autoqa.koji_utils as koji_utils
import autoqa.test as test
import autoqa.bodhi_utils as bodhi_utils
from autoqa.repoinfo import repoinfo
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autoqa.test import TestDetail

class upgradepath(AutoQATest):
    version = 1 # increment if setup() changes


    @ExceptionCatcher()
    def initialize(self, config, event, kojitag, **kwargs):
        super(self.__class__, self).initialize(config)

        # fill in self.detail.id
        if event == "post-bodhi-update-batch":
            self.detail.id = kojitag + '-pending'
        else: # event is post-bodhi-update
            self.detail.id = kwargs['name']

        # mapping of package build -> TestDetail for the corresponding Bodhi update
        self.build2detail = {}

        # dict of problematic builds (Koji/Bodhi problems): {ENVR: error}
        self.problematic = {}

        self.koji = koji_utils.SimpleKojiClientSession()
        self.koji.check_connection()


    def compare(self, proposed_build, tags, op, test_detail):
        '''Compare proposed_build with given operator to latest build in
        all tags and update test_detail.result.
        @param proposed_build Koji Build object
        @param tags list of kojitags - a string or a tuple of strings. if it is
                    a tuple, then the contents of corresponding repos will be
                    united while searching for latest build
        @param op comparison operator from module 'operator'
        @param test_detail TestDetail object related to the proposed build
        '''
        result = 'PASSED'

        # convert 'op' to human language for error message clarity
        opstr = {operator.eq: 'equal to',
                 operator.ge: 'greater than or equal to',
                 operator.gt: 'greater than',
                 operator.le: 'less than or equal to',
                 operator.lt: 'less than',
                 operator.ne: 'not equal to'}
        assert op in opstr, 'Used unsupported operator: %s' % op

        # test the build against all proposed tags or set of tags
        for tag_tuple in tags:
            # convert everything to tuples
            if isinstance(tag_tuple, str):
                tag_tuple = (tag_tuple,)
            # now find the latest build in repositories
            latest_build = None
            for tag in tag_tuple:
                build = self.koji.latest_by_tag(tag, proposed_build['name'])
                if latest_build is None or (build is not None and
                    koji_utils.compareEVR(build, latest_build) > 0):
                    latest_build = build
            # and make human readable string for this tag tuple
            tag_str = ' + '.join([tag for tag in tag_tuple])

            if latest_build is None:
                # no such package in repos means everything is fine
                self.log('{0:<7}{1}'.format('[ OK ]', tag_str),
                    details=test_detail)
                self.log('\tNo such package: %s' % proposed_build['name'],
                    details=test_detail)
            else:
                res = koji_utils.compareEVR(proposed_build, latest_build)

                if op(res, 0):
                    ok = True
                    self.log('{0:<7}{1}'.format('[ OK ]', tag_str),
                        details=test_detail)
                else:
                    ok = False
                    self.log('{0:<7}{1}'.format('[FAIL]', tag_str),
                        details=test_detail, highlight=True)
                self.log('\tLatest package: %s' % koji_utils.getENVR(latest_build),
                    details=test_detail)

                if not ok:
                    self.log('\tError: Proposed package must be %s the latest '
                        'package' % opstr[op], details=test_detail)
                    result = 'FAILED'

        test_detail.update_result(result)


    @ExceptionCatcher()
    def run_once(self, kojitag, event, **kwargs):
        super(self.__class__, self).run_once()

        # Ideally upgradepath should check just the new updates. But since we don't
        # yet support test re-scheduling [1], we have to work around somehow [2].
        # So let's just run upgradepath for *all* updates requesting their move to
        # <kojitag>. It is a little inefficient, but if for some package the result
        # changes, we will report it correctly (and that's more important).
        # When test re-scheduling is supported, we can revert this test from whole
        # tag testing to just new updates testing.
        # [1] https://fedorahosted.org/autoqa/ticket/245
        # [2] https://fedorahosted.org/autoqa/ticket/246

        # THE UPGRADEPATH ALGORITHM:
        #
        # == Pushing to main repository ==
        # Pushing PKG to F(N)-main means:
        #   1. PKG in F(lower)-main <= PKG to push
        #   2. PKG in F(higher)-main >= PKG to push
        #
        # == Pushing to updates repository ==
        # Pushing PKG to F(N)-updates means:
        #   1. PKG in F(lower)-main <= PKG to push
        #   2. PKG in F(lower)-updates <= PKG to push
        #   3. PKG in F(higher)-main union F(higher)-updates => PKG to push
        #
        # Note: If PKG doesn't exist in REPO, it also satisfies any condition
        #
        # References:
        #   https://fedorahosted.org/autoqa/ticket/309
        #   https://fedorahosted.org/autoqa/ticket/277
        #
        # FIXME: Branched releases before "Final Change Deadline" use -updates
        # tag for pushing into -main repository. That breaks this algorithm.
        # See: https://fedorahosted.org/autoqa/ticket/277#comment:2
        # For proper functionality we need to detect such branched releases and
        # adjust the algorithm accordingly (run "Pushing to main repository" part
        # even if the tag does not correspond to it).

        # Get a list of all repos we monitor (currently not -testing)
        # FIXME - perhaps we should only query for 'active' repos
        repos = [repoinfo.getrepo(reponame) for reponame in repoinfo.repos()
                 if not reponame.endswith('-testing')]
        assert kojitag in [repo['tag'] for repo in repos], 'Requested unsupported kojitag: %s' % kojitag

        # are we pushing to main repo or to updates repo?
        # we recognize main repository when it has no parents
        push_to_main = not bool(repoinfo.getrepo_by_tag(kojitag)['parents'])

        if push_to_main:
            # we need to get list of main repositories for F(higher) and F(lower)
            low_repos = [repo for repo in repos if not repo['parents'] and repo['tag'] < kojitag]
            hi_repos = [repo for repo in repos if not repo['parents'] and repo['tag'] > kojitag]
        else: # pushing to update repo
            # for both F(lower) and F(higher) we are interested in main and update repos
            low_repos = [repo for repo in repos if repo['tag'] < kojitag]
            hi_repos = [repo for repo in repos if repo['tag'] > kojitag]
        reposorter = lambda repo : repo['name']
        low_repos = sorted(low_repos, key=reposorter)
        hi_repos = sorted(hi_repos, key=reposorter)

        # gather list of builds to test
        if event == "post-bodhi-update-batch":
            # get the list of all builds requesting move to the kojitag
            sourcetag = kojitag + '-pending'
            self.log('Koji tag to be tested: %s' % sourcetag)
            self.log('Querying Koji for builds...')
            updates = self.koji.listTagged(sourcetag)
            envrs = sorted([koji_utils.getENVR(u) for u in updates])
        else: # event is post-bodhi-update
            envrs = kwargs['nvrs']
        self.log('Builds to be tested:\n  %s' % '\n  '.join(envrs))

        # For every build tested get a Bodhi update that corresponds to it.
        # Check that all builds for that update are scheduled for testing.
        # If everything is ok, put that into a dictionary:
        #
        # build2update = {build NVR: BodhiUpdate}
        #
        # If there is some problem, put that build into self.problematic and
        # remove it from 'envrs'.

        self.log('Querying Bodhi for matching updates...')
        build2update, failures = bodhi_utils.build2update(envrs, strict=True)

        for build, update in failures.items():
            if update is None:
                # build not found in Bodhi
                msg = 'Update containing %s not found in Bodhi' % build
            else:
                # some other builds from the matching update were not scheduled
                msg = ("Build %s is part of an incompletely scheduled Bodhi "
                       "update '%s'." % (build, update['title']))
            self.log(msg, highlight=True, stderr=True)
            self.problematic[build] = msg
            envrs.remove(build)

        # we need Bodhi object to be hashable to use them in sets and dicts
        # let's convert it
        bodhi_objects = set()
        for build, update in build2update.items():
            same = [bo for bo in bodhi_objects if bo.update is update]
            if same:
                build2update[build] = same[0]
            else:
                build2update[build] = BodhiUpdate(update)
                bodhi_objects.add(build2update[build])

        self.log('Bodhi updates found:\n  %s' % '\n  '.join(
            sorted([update.update['title'] for update in
                    set(build2update.values())])))


        # Now it is time to populate self.build2detail:
        # {build NVR: TestDetail}
        # All builds from the same Bodhi update should get the same
        # TestDetail object.
        #
        # As an intermediate step use update2detail variable:
        # {BodhiUpdate: TestDetail}

        update2detail = {}
        for update in set(build2update.values()):
            update2detail[update] = TestDetail(test=self,
                                               id=update.update['title'])

        for build, update in build2update.iteritems():
            self.build2detail[build] = update2detail[update]


        # Let the testing begin!
        self.log('Running main upgradepath tests...')
        for envr in envrs:
            detail = self.build2detail[envr]
            if detail.broken():
                # this Bodhi update was already broken by some envr, skip it
                continue

            self.log('%s\n%s into %s\n%s' % (60*'=', envr, kojitag, 60*'='),
                details=detail)

            # get all info about this current build
            proposed_build = self.koji.getBuild(envr)
            if proposed_build is None:
                msg = 'This build does not exist in Koji: %s' % envr
                self.log(msg, highlight=True, stderr=True, details=detail)
                self.problematic[envr] = msg
                detail.update_result('ABORTED')
                continue

            # if we want to push an older build, warn about that and perform further
            # checking for the most recent one
            # see https://fedorahosted.org/autoqa/ticket/230
            latest_build = self.koji.latest_by_tag(kojitag, proposed_build['name'])
            if latest_build is not None:
                if koji_utils.compareEVR(proposed_build, latest_build) <= 0:
                    msg = ('Warning: Same or newer build already exists in the '
                           'same repo.\nAll checks will be performed for: %s' %
                           koji_utils.getENVR(latest_build))
                    self.log(msg, highlight=True, stderr=True, details=detail)
                    proposed_build = latest_build

            if push_to_main:
                # compare with lower tags, so version has to be greater or equal
                low_tags = [repo['tag'] for repo in low_repos]
                self.compare(proposed_build, low_tags, operator.ge, detail)

                # compare with higher tags, so version has to be lower or equal
                hi_tags = [repo['tag'] for repo in hi_repos]
                self.compare(proposed_build, hi_tags, operator.le, detail)

            else: # pushing to update repository
                # compare with lower tags, so version has to be greater or equal
                low_tags = [repo['tag'] for repo in low_repos]
                self.compare(proposed_build, low_tags, operator.ge, detail)

                # compare with higher tags, so version has to be lower or equal
                # we have to group corresponding main and update repos
                # let's define dict parent -> [child1, child2]
                # e.g. 'dist-f14' -> ['dist-f14-updates', 'dist-f14-updates-candidate']
                tag_family = {}
                for repo in hi_repos:
                    # this presumes all repos are sorted by name/tag
                    if not bool(repo['parents']):
                        tag_family[repo['tag']] = []
                    else:
                        main_parent = repoinfo.getrepo(repo['parents'][0])
                        if main_parent['tag'] in tag_family:
                            tag_family[main_parent['tag']].append(repo['tag'])
                        else:
                            # it might happen that there is no main_parent for
                            # a repo. e.g. we want to push to dist-f14 (so we don't
                            # check this tag), but we still want to check dist-f14-updates
                            # in this case just add it as a key
                            tag_family[repo['tag']] = []

                hi_tags = []
                for family in sorted(tag_family):
                    tag_tuple = [family]
                    tag_tuple.extend(tag_family[family])
                    hi_tags.append(tag_tuple)

                self.compare(proposed_build, hi_tags, operator.le, detail)

            self.log('RESULT: %s' % detail.result, details=detail)

        # after running all tests, update self.detail.result
        for detail in set(self.build2detail.values()):
            self.detail.update_result(detail.result)
        # if some broken package, end the test as NEEDS_INSPECTION
        if self.problematic:
            self.log('Broken updates found, setting final result to '
                'NEEDS_INSPECTION', stderr=True)
            self.detail.update_result('NEEDS_INSPECTION')

        # create summary
        summary = test.create_multi_item_summary(set(self.build2detail.values()))
        self.detail.summary = summary
        self.log('SUMMARY: ' + self.detail.summary)

        # report results
        self.log('Submitting all test results...')
        for update in sorted(set(build2update.values()),
                             key=lambda u: u.update['title']):
            detail = update2detail[update]
            builds = [build for build in build2update.keys()
                      if build2update[build] == update]

            # create {result: [build]}
            results = {}
            for build in builds:
                result = self.build2detail[build].result
                results.setdefault(result, [])
                results[result].append(build)

            # add additional log information
            detail.log.summary_detail = results
            detail.log.kojitag = kojitag + '-pending'
            if update.update['updateid']:
                detail.log.bodhi_id = update.update['updateid']

            # submit results
            self.post_results(detail, bodhi={'title': update.update['title']})


class BodhiUpdate(object):
    '''This a custom object wrapper for Bodhi update object. The purpose of this
    class is to make Bodhi updates hashable (therefore objects, the original ones
    are dictionaries), so that we can create dictionaries with Bodhi updates as
    their keys.'''
    def __init__(self, update):
        self.update = update
